@@ -41,10 +41,12 @@ This module also provides a utility method:
 """
 
 import copy
+from typing import List, Union
 
 from parlai.core.build_data import modelzoo_path
 from parlai.core.loader import load_agent_module
 from parlai.core.loader import register_agent  # noqa: F401
+from parlai.core.message import Message
 from parlai.core.opt import Opt
 from parlai.utils.misc import warn_once
 import parlai.utils.logging as logging
@@ -154,6 +156,70 @@ class Agent(object):
         """
         pass
 
+    def respond(
+        self, text_or_message: Union[str, Message], **other_message_fields
+    ) -> str:
+        """
+        An agent convenience function which calls the act() and provides a string
+        response to a text or message field.
+
+        :param Union[str, Message] text_or_message:
+            A string for the 'text' field or a message which MUST
+            comprise of the 'text' field apart from other fields.
+        :param kwargs other_message_fields:
+            Provide fields for the message in the form of keyowrd arguments.
+        :return:
+            Agent's response to the message.
+        :rtype:
+            str
+        """
+        if isinstance(text_or_message, str):
+            observation = Message(text=text_or_message, **other_message_fields)
+        else:
+            observation = Message(**text_or_message, **other_message_fields)
+            if 'text' not in observation:
+                raise RuntimeError('The agent needs a \'text\' field in the message.')
+
+        if 'episode_done' not in observation:
+            observation['episode_done'] = True
+        agent = self.clone()
+        agent.observe(observation)
+        response = agent.act()
+        return response['text']
+
+    def batch_respond(self, messages: List[Message]) -> List[str]:
+        """
+        An agent convenience function which calls the batch_act() and provides a batch
+        response to a list of messages.
+
+        :param List[Message] messages:
+            A list of messages each of which MUST comprise of the 'text' field
+            apart from other fields.
+        :return:
+            Agent's batch response to the messages.
+        :rtype:
+            List[str]
+        """
+        observations = []
+        agents = []
+        for i, message in enumerate(messages):
+            if 'text' not in message:
+                raise RuntimeError(
+                    'The agent needs a \'text\' field in the {}th message.'.format(i)
+                )
+            if 'episode_done' not in message:
+                message['episode_done'] = True
+            agent = self.clone()
+            agents.append(agent)
+            observations.append(agent.observe(message))
+        agent_acts = self.batch_act(observations)
+        response = []
+        for agent, resp in zip(agents, agent_acts):
+            if hasattr(agent, "self_observe"):
+                agent.self_observe(resp)
+            response.append(resp['text'])
+        return response
+
     @classmethod
     def upgrade_opt(cls, opt_from_disk: Opt):
         """
@@ -245,7 +311,7 @@ def compare_init_model_opts(opt: Opt, curr_opt: Opt):
     # print warnings
     extra_strs = ['{}: {}'.format(k, v) for k, v in extra_opts.items()]
     if extra_strs:
-        logging.warn(
+        logging.warning(
             'your model is being loaded with opts that do not '
             'exist in the model you are initializing the weights with: '
             '{}'.format(','.join(extra_strs))
@@ -255,7 +321,7 @@ def compare_init_model_opts(opt: Opt, curr_opt: Opt):
         '--{} {}'.format(k.replace('_', '-'), v) for k, v in different_opts.items()
     ]
     if different_strs:
-        logging.warn(
+        logging.warning(
             'your model is being loaded with opts that differ '
             'from the model you are initializing the weights with. Add the '
             'following args to your run command to change this: \n'
@@ -309,7 +375,7 @@ def create_agent_from_opt_file(opt: Opt):
     if opt.get('override'):
         for k, v in opt['override'].items():
             if k in opt_from_file and str(v) != str(opt_from_file.get(k)):
-                logging.warn(
+                logging.warning(
                     f'Overriding opt["{k}"] to {v} (previously: {opt_from_file.get(k)})'
                 )
             opt_from_file[k] = v
@@ -324,7 +390,13 @@ def create_agent_from_opt_file(opt: Opt):
         if k not in opt_from_file:
             opt_from_file[k] = v
 
-    opt_from_file['model_file'] = model_file  # update model file path
+    # update model file path to the one set by opt
+    opt_from_file['model_file'] = model_file
+    # update init model path to the one set by opt
+    # NOTE: this step is necessary when for example the 'init_model' is
+    # set by the Train Loop (as is the case when loading from checkpoint)
+    if opt.get('init_model') is not None:
+        opt_from_file['init_model'] = opt['init_model']
 
     # update dict file path
     if not opt_from_file.get('dict_file'):
@@ -358,7 +430,7 @@ def add_datapath_and_model_args(opt: Opt):
     # add model args if they are missing
     model = get_model_name(opt)
     if model is not None:
-        parser.add_model_subargs(model)
+        parser.add_model_subargs(model, opt)
     opt_parser = parser.parse_args("")
     for k, v in opt_parser.items():
         if k not in opt:
@@ -407,7 +479,7 @@ def create_agent(opt: Opt, requireModelExists=False):
         model = model_class(opt)
         if requireModelExists and hasattr(model, 'load') and not opt.get('model_file'):
             # double check that we didn't forget to set model_file on loadable model
-            logging.warn('model_file unset but model has a `load` function.')
+            logging.warning('model_file unset but model has a `load` function.')
         return model
     else:
         raise RuntimeError('Need to set `model` argument to use create_agent.')
