@@ -8,10 +8,14 @@
 Opt is the system for passing around options throughout ParlAI.
 """
 
+from __future__ import annotations
+
 import copy
 import json
 import pickle
 import traceback
+import os
+import pkg_resources
 import parlai.utils.logging as logging
 
 from typing import List
@@ -25,9 +29,10 @@ __AUTOCLEAN_KEYS__: List[str] = [
     "batchindex",
     "download_path",
     "datapath",
-    "batchindex",
-    # we don't save interactive mode, it's only decided by scripts or CLI
+    "verbose",
+    # we don't save interactive mode or load from checkpoint, it's only decided by scripts or CLI
     "interactive_mode",
+    "load_from_checkpoint",
 ]
 
 
@@ -45,7 +50,7 @@ class Opt(dict):
         self.deepcopies = []
 
     def __setitem__(self, key, val):
-        loc = traceback.format_stack()[-2]
+        loc = traceback.format_stack(limit=2)[-2]
         self.history.append((key, val, loc))
         super().__setitem__(key, val)
 
@@ -64,7 +69,7 @@ class Opt(dict):
         Override deepcopy so that history is copied over to new object.
         """
         # track location of deepcopy
-        loc = traceback.format_stack()[-3]
+        loc = traceback.format_stack(limit=3)[-3]
         self.deepcopies.append(loc)
         # copy all our children
         memo = Opt({k: copy.deepcopy(v) for k, v in self.items()})
@@ -118,7 +123,7 @@ class Opt(dict):
             f.write('\n')
 
     @classmethod
-    def load(cls, optfile: str) -> 'Opt':
+    def load(cls, optfile: str) -> Opt:
         """
         Load an Opt from disk.
         """
@@ -135,6 +140,45 @@ class Opt(dict):
                 del dct[key]
         return cls(dct)
 
+    @classmethod
+    def load_init(cls, optfile: str) -> Opt:
+        """
+        Like load, but also looks in opt_presets folders.
+
+        optfile may also be a comma-separated list of multiple presets/files.
+        """
+        if "," in optfile:
+            # load and combine each of the individual files
+            new_opt = cls()
+            for subopt in optfile.split(","):
+                new_opt.update(cls.load_init(subopt))
+            return new_opt
+
+        oa_filename = os.path.join("opt_presets", optfile + ".opt")
+        user_filename = os.path.join(os.path.expanduser(f"~/.parlai"), oa_filename)
+        if PathManager.exists(optfile):
+            return cls.load(optfile)
+        elif PathManager.exists(user_filename):
+            # use a user's custom opt preset
+            return cls.load(user_filename)
+        else:
+            # Maybe a bundled opt preset
+            for root in ['parlai', 'parlai_internal', 'parlai_fb']:
+                try:
+                    if pkg_resources.resource_exists(root, oa_filename):
+                        return cls.load(
+                            pkg_resources.resource_filename(root, oa_filename)
+                        )
+                except ModuleNotFoundError:
+                    continue
+
+        # made it through without a return path so raise the error
+        raise FileNotFoundError(
+            f"Could not find filename '{optfile} or opt preset '{optfile}.opt'. "
+            "Please check https://parl.ai/docs/opt_presets.html for a list "
+            "of available opt presets."
+        )
+
     def log(self, header="Opt"):
         from parlai.core.params import print_git_commit
 
