@@ -19,14 +19,14 @@ all TorchRankerAgents and TorchGeneratorAgents support this.
 ## Examples
 
 ```shell
-parlai multiprocessing_train -m transformer/generator -bs 16 -t convai2 -mf /tmp/mymodel
+parlai multiprocessing_train -m transformer/generator --batchsize 16 --task convai2 --model-file /tmp/mymodel
 ```
 """
 
 import torch
-import random
 import os
 import signal
+import traceback
 import parlai.scripts.train_model as single_train
 import parlai.utils.distributed as distributed_utils
 from parlai.core.script import ParlaiScript, register_script
@@ -35,26 +35,40 @@ from parlai.core.script import ParlaiScript, register_script
 def multiprocess_train(
     rank, opt, port=61337, rank_offset=0, gpu=None, hostname='localhost'
 ):
+    init_method = f"tcp://{hostname}:{port}"
     with distributed_utils.distributed_context(
-        rank, opt, port, rank_offset, gpu, hostname
+        rank, opt, rank_offset, gpu, init_method=init_method
     ) as opt:
         # Run the actual training
         opt['multiprocessing'] = True
-        return single_train.TrainLoop(opt).train()
+        try:
+            return single_train.TrainLoop(opt).train()
+        except Exception:
+            import parlai.utils.logging as logging
+
+            logging.critical(traceback.format_exc())
+            logging.critical(
+                f"Got the above exception on worker {rank + rank_offset}. "
+                "This may cause hangs requiring manual killing of processes."
+            )
+            raise
 
 
-def launch_and_train(opt, port):
+def launch_and_train(opt, port=None):
     """
     Perform a fork() to many processes.
     """
+    if port is None:
+        port = distributed_utils.find_free_port()
     # Launch multiple subprocesses
-    spawncontext = torch.multiprocessing.spawn(
+    spawncontext = torch.multiprocessing.start_processes(
         multiprocess_train,
         # need to give rank offset as 1 to cover the fact that the main
         # process is rank 0, but that spawn() doesn't let you control rank
         (opt, port, 1),
         nprocs=opt['distributed_world_size'] - 1,  # main proc will also run loop
         join=False,
+        start_method='spawn',
     )
 
     try:
@@ -80,10 +94,15 @@ def setup_args():
 class MultiProcessTrain(ParlaiScript):
     @classmethod
     def setup_args(cls):
-        return setup_args()
+        argparser = setup_args()
+        argparser.add_argument('--port', type=int, default=None)
+        return argparser
 
     def run(self):
-        port = random.randint(32000, 48000)
+        if self.opt['port'] is None:
+            port = None
+        else:
+            port = self.opt['port']
         return launch_and_train(self.opt, port)
 
 
