@@ -26,10 +26,10 @@ def setup_args(parser=None):
     parser.add_argument('--seed', type=int, default=42)
     parser.add_argument('-d', '--display-examples', type='bool', default=True)
     parser.add_argument(
-        '--display-ignore-fields',
+        '--display-add-fields',
         type=str,
-        default='label_candidates,text_candidates',
-        help='Do not display these fields',
+        default='',
+        help='Display these fields when verbose is off (e.g., "--display-add-fields label_candidates,beam_texts")',
     )
     parser.add_argument(
         '-st',
@@ -49,9 +49,15 @@ def setup_args(parser=None):
     )
     parser.add_argument(
         '--seed-messages-from-task',
-        action='store_true',
+        type='bool',
+        default=False,
         help='Automatically seed conversation with messages from task dataset.',
     )
+    parser.add_argument(
+        '--seed-messages-from-file',
+        default=None,
+        help='If specified, loads newline-separated strings from the file as conversation starters.',
+    )
     parser.add_argument(
         '--outfile', type=str, default=None, help='File to save self chat logs'
     )
@@ -74,14 +80,14 @@ def setup_args(parser=None):
         help='Path to file containing opts to override for partner',
     )
     parser.set_defaults(interactive_mode=True, task='self_chat')
-    WorldLogger.add_cmdline_args(parser)
+    WorldLogger.add_cmdline_args(parser, partial_opt=None)
     return parser
 
 
 def _run_self_chat_episode(opt, world, world_logger):
     bsz = opt.get('batchsize', 1)
     num_turns = opt['selfchat_max_turns']
-
+    assert bsz == 1, "Batch size cannot be different than 1 for self-chat"
     num_parleys = math.ceil(num_turns / bsz)
     for _ in range(num_parleys):
         world.parley()
