@@ -8,7 +8,7 @@ import copy
 import random
 from typing import Any, Dict, List, Optional
 
-from parlai.agents.repeat_label.repeat_label import RepeatLabelAgent
+from parlai.agents.fixed_response.fixed_response import FixedResponseAgent
 from parlai.core.agents import Agent
 from parlai.core.worlds import create_task, DialogPartnerWorld, validate
 from parlai.core.message import Message
@@ -31,7 +31,8 @@ def load_openers(opt) -> Optional[List[str]]:
         task_opt['datatype'] = f'{datatype}:evalmode'
     task_opt['interactive_task'] = False
     task_opt['selfchat_task'] = False
-    task_agent = RepeatLabelAgent(task_opt)
+    task_opt['fixed_response'] = None
+    task_agent = FixedResponseAgent(task_opt)
     task_world = create_task(task_opt, task_agent)
 
     # run through task data, collecting all first messages
@@ -49,6 +50,13 @@ def load_openers(opt) -> Optional[List[str]]:
     return list(openers)
 
 
+def load_openers_from_file(filepath: str) -> List[str]:
+    openers = []
+    with open(filepath, 'r') as f:
+        openers = [l.strip() for l in f]
+    return openers
+
+
 class SelfChatWorld(DialogPartnerWorld):
     def __init__(self, opt, agents, shared=None):
         super().__init__(opt, agents, shared)
@@ -80,6 +88,8 @@ class SelfChatWorld(DialogPartnerWorld):
         """
         if self.opt.get('seed_messages_from_task'):
             self._openers = load_openers(self.opt)
+        elif self.opt.get('seed_messages_from_file'):
+            self._openers = load_openers_from_file(self.opt['seed_messages_from_file'])
 
     def get_openers(self, episode_num: int) -> Optional[List[str]]:
         """
@@ -153,6 +163,7 @@ class SelfChatWorld(DialogPartnerWorld):
                 if len(utts) > i:
                     self.acts[i] = utts[i]
                     if hasattr(self.agents[i], 'self_observe'):
+                        self.agents[i].observe({'episode_done': False})
                         self.agents[i].self_observe(self.acts[i])
                 else:
                     self.acts[i] = self.agents[i].act()
