@@ -15,9 +15,17 @@ for unseen) after the last colon in the task.
 E.g. `wizard_of_wikipedia:WizardDialogKnowledgeTeacher:random_split`
 """
 
+from __future__ import annotations
+from typing import Iterable, Optional, Tuple
+from parlai.core.message import Message
+from parlai.core.metrics import AverageMetric, normalize_answer, F1Metric
+from parlai.core.params import ParlaiParser
+from parlai.core.opt import Opt
 import copy
 from parlai.core.teachers import FixedDialogTeacher, MultiTaskTeacher
 from parlai.utils.io import PathManager
+from parlai.utils import logging
+from parlai.utils.misc import warn_once
 from .build import build
 
 import json
@@ -95,6 +103,73 @@ def _path(opt, split='random_split'):
     return os.path.join(dp, df)
 
 
+class RareWordF1Calculator:
+    """
+    Helper class for computing F1 with an emphasis on infrequent words.
+    """
+
+    def __init__(self, corpus: str, top_p: float = 0.5):
+        try:
+            import nltk
+        except ImportError:
+            raise ImportError('Please install nltk (e.g. pip install nltk).')
+        words = normalize_answer(corpus).split()
+        self._freq_dist = nltk.FreqDist(words)
+        self._cutoff_count = RareWordF1Calculator._find_cutoff_count(
+            self._freq_dist, top_p
+        )
+
+    @property
+    def freq_dist(self):
+        return self._freq_dist
+
+    @staticmethod
+    def _find_cutoff_count(freq_dist, top_p: float) -> int:
+        """
+        Finds the word occurance for which the cumulative occurances are `top_p` of the
+        overall word count.
+        """
+        assert top_p < 1
+        target = sum(freq_dist.values()) * top_p
+        cumul = 0
+        for _, v in freq_dist.most_common():
+            cumul += v
+            if cumul > target:
+                return v
+        raise RuntimeError(f"Invalid top {top_p*100}% of the corpus distribution")
+
+    @staticmethod
+    def _filter(freq_dist, cutoff: int, text: str) -> str:
+        """
+        For words that are found in the reference distribution, filters those with an
+        occurrence count less than the cutoff.
+        """
+        words = normalize_answer(text).split()
+        return " ".join([w for w in words if freq_dist.get(w, cutoff) < cutoff])
+
+    def compute(self, guess: str, answers: Iterable[str]) -> F1Metric:
+        if guess is None or answers is None:
+            return F1Metric(0, 0)
+        guess = RareWordF1Calculator._filter(self._freq_dist, self._cutoff_count, guess)
+        answers = [
+            RareWordF1Calculator._filter(self._freq_dist, self._cutoff_count, a)
+            for a in answers
+        ]
+        if not any(len(a) for a in answers):
+            # no rare words in labels, set denominator to zero
+            return F1Metric(0, 0)
+        return F1Metric.compute(guess, answers)
+
+
+def _build_rare_word_f1(datapath: str) -> RareWordF1Calculator:
+    all_text = ''
+    data_path = os.path.join(datapath, 'wizard_of_wikipedia', 'data.json')
+    with PathManager.open(data_path) as f:
+        data = json.load(f)
+        all_text += ' '.join(m['text'] for d in data for m in d['dialog']) + ' '
+    return RareWordF1Calculator(all_text, top_p=0.5)
+
+
 class WizardOfWikipediaTeacher(FixedDialogTeacher):
     """
     The default teacher; essentially reads the json file and outputs the raw data.
@@ -198,6 +273,7 @@ class WizardDialogKnowledgeTeacher(WizardOfWikipediaTeacher):
     """
 
     def __init__(self, opt, shared=None):
+        self.add_missing_turns = opt.get('add_missing_turns', 'none')
         super().__init__(opt, shared)
         self.label_type = opt.get('label_type', 'response')
         self.include_knowledge = opt.get('include_knowledge', True)
@@ -205,10 +281,17 @@ class WizardDialogKnowledgeTeacher(WizardOfWikipediaTeacher):
         self.knowledge_separator = opt.get('include_knowledge_separator', False)
         self.chosen_topic_delimiter = opt.get('chosen_topic_delimiter', '\n')
         self.num_exs = sum(self.len_episode(i) for i in range(len(self.data)))
+        if shared and 'rare_word_f1' in shared:
+            self.rare_word_f1 = shared['rare_word_f1']
+        elif self.label_type == 'response':
+            self.rare_word_f1 = _build_rare_word_f1(opt['datapath'])
 
-    @staticmethod
-    def add_cmdline_args(argparser):
-        agent = argparser.add_argument_group('Wizard Dialog Knowledge arguments')
+    @classmethod
+    def add_cmdline_args(
+        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
+    ) -> ParlaiParser:
+        super().add_cmdline_args(parser, partial_opt)
+        agent = parser.add_argument_group('Wizard Dialog Knowledge arguments')
         agent.add_argument(
             '--label-type',
             type=str,
@@ -248,12 +331,39 @@ class WizardDialogKnowledgeTeacher(WizardOfWikipediaTeacher):
             help='in interactive mode, this is the number of topic choices'
             'the human will have',
         )
+        agent.add_argument(
+            '--add-missing-turns',
+            type=str,
+            choices=['none', 'train', 'all'],
+            default='none',
+            help='For reproducibility, the default "none" is the previous version which misssing some data.'
+            'When "train" is chosen, only the training set is supplemented.'
+            'When "all" is chosen, all data are supplemented.',
+        )
+        return parser
+
+    def share(self):
+        shared = super().share()
+        if hasattr(self, 'rare_word_f1'):
+            shared['rare_word_f1'] = self.rare_word_f1
+        return shared
 
     def len_episode(self, ep):
         d = self.data[ep]
         wizard_first = 'Wizard' in d['dialog'][0]['speaker']
         if wizard_first:
-            return (len(d['dialog']) - 1) // 2
+            if self.add_missing_turns == 'none':
+                warn_once(
+                    'Some data not being used. If you are not trying to reproduce '
+                    'the previous results, it is recommended that you run with the '
+                    'flag --add-missing-turns train or --add-missing-turns all.'
+                )
+                len_ep = (len(d['dialog']) - 1) // 2
+            elif self.add_missing_turns == 'train' and 'train' not in self.datatype:
+                len_ep = (len(d['dialog']) - 1) // 2
+            else:
+                len_ep = (len(d['dialog']) - 1) // 2 + 1
+            return len_ep
         return len(d['dialog']) // 2
 
     def num_examples(self):
@@ -347,6 +457,86 @@ class WizardDialogKnowledgeTeacher(WizardOfWikipediaTeacher):
             action['checked_sentence'] = sentence
         return action
 
+    def custom_evaluation(
+        self,
+        teacher_action: Message,
+        labels: Optional[Tuple[str]],
+        model_response: Message,
+    ):
+        """
+        Custom Evaluations for Wizard of Wikipedia.
+
+        When the label is `chosen_sent`, evaluate whether the model response...
+        1) Is the correct document (title)
+        2) _contains_ the correct chosen sentence (even if it's not wholly the answer)
+
+        When the label is `response`, we compute F1 of model generation w.r.t checked sentence.
+
+        :param teacher_action:
+            The message last sent from this teacher.
+        :param labels:
+            The previous correct labels, if there were any.
+        :param model_response:
+            The raw response from the model. Generally you want to rely on the
+            text field, but others may be necessary in specific situations.
+        """
+        if (
+            self.label_type == 'response'
+            and 'text' in model_response
+            and 'checked_sentence' in teacher_action
+        ):
+            self.metrics.add(
+                'knowledge_f1',
+                F1Metric.compute(
+                    model_response['text'], [teacher_action['checked_sentence']]
+                ),
+            )
+            if labels:
+                self.metrics.add(
+                    'rare_word_f1',
+                    self.rare_word_f1.compute(model_response['text'], labels),
+                )
+        elif (
+            self.label_type == 'chosen_sent'
+            and TOKEN_KNOWLEDGE in model_response['text']
+        ):
+            try:
+                correct_title, correct_passage = [
+                    normalize_answer(a) for a in labels[0].split(TOKEN_KNOWLEDGE)
+                ]
+            except ValueError:
+                # Knowledge not chosen
+                correct_title, correct_passage = TOKEN_NOCHOSEN, TOKEN_NOCHOSEN
+            title, passage = [
+                normalize_answer(a)
+                for a in model_response['text'].split(TOKEN_KNOWLEDGE)
+            ]
+
+            self.metrics.add('title_r@1', AverageMetric(int(correct_title == title)))
+            self.metrics.add(
+                'passage_r@1', AverageMetric(int(correct_passage in passage))
+            )
+            if 'title_candidates' in model_response:
+                title_candidates = [
+                    normalize_answer(t) for t in model_response['title_candidates']
+                ][:5]
+                self.metrics.add(
+                    'title_r@5',
+                    AverageMetric(
+                        int(any(correct_title == t for t in title_candidates))
+                    ),
+                )
+            if 'text_candidates' in model_response:
+                text_candidates = [
+                    normalize_answer(t) for t in model_response['text_candidates']
+                ][:5]
+                self.metrics.add(
+                    'passage_r@5',
+                    AverageMetric(
+                        int(any(correct_passage in t for t in text_candidates))
+                    ),
+                )
+
 
 class BasicdialogTeacher(WizardOfWikipediaTeacher):
     """
@@ -354,14 +544,18 @@ class BasicdialogTeacher(WizardOfWikipediaTeacher):
     """
 
     def __init__(self, opt, shared=None):
+        self.add_missing_turns = opt.get('add_missing_turns', 'none')
         super().__init__(opt, shared)
         self.speaker_label = opt.get('speaker_label', 'both')
         self.add_topic = opt.get('add_topic', False)
         self.num_exs = sum(self.len_episode(i) for i in range(len(self.data)))
 
-    @staticmethod
-    def add_cmdline_args(argparser):
-        agent = argparser.add_argument_group('Basic Dialog Arguments')
+    @classmethod
+    def add_cmdline_args(
+        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
+    ) -> ParlaiParser:
+        super().add_cmdline_args(parser, partial_opt)
+        agent = parser.add_argument_group('Basic Dialog Arguments')
         agent.add_argument(
             '--speaker-label',
             type=str,
@@ -375,6 +569,16 @@ class BasicdialogTeacher(WizardOfWikipediaTeacher):
             default=False,
             help='prepend chosen topic to first turn',
         )
+        agent.add_argument(
+            '--add-missing-turns',
+            type=str,
+            choices=['none', 'train', 'all'],
+            default='none',
+            help='For reproducibility, the default "none" is the previous version which missing some data. '
+            'When "train" is chosen, only the training set is supplemented. '
+            'When "all" is chosen, all data are supplemented.',
+        )
+        return parser
 
     def num_examples(self):
         return self.num_exs
@@ -383,7 +587,18 @@ class BasicdialogTeacher(WizardOfWikipediaTeacher):
         d = self.data[ep]
         first_speaker = d['dialog'][0]['speaker'].lower()
         if self.speaker_label != 'both' and self.speaker_label in first_speaker:
-            return (len(d['dialog']) - 1) // 2
+            if self.add_missing_turns == 'none':
+                warn_once(
+                    'Some data not being used. If you are not trying to reproduce '
+                    'the previous results, it is recommended that you run with the '
+                    'flag --add-missing-turns train or --add-missing-turns all.'
+                )
+                len_ep = (len(d['dialog']) - 1) // 2
+            elif self.add_missing_turns == 'train' and 'train' not in self.datatype:
+                len_ep = (len(d['dialog']) - 1) // 2
+            else:
+                len_ep = (len(d['dialog']) - 1) // 2 + 1
+            return len_ep
         return len(d['dialog']) // 2
 
     def get(self, episode_idx, entry_idx=0):
@@ -466,11 +681,13 @@ class GeneratorTeacher(WizardDialogKnowledgeTeacher):
         self.gold_knowledge_delimiter = opt.get('gold_knowledge_delimiter', '\n')
         self.dropout = opt.get('ignorant_dropout', 0.0)
 
-    @staticmethod
-    def add_cmdline_args(argparser):
-        argparser.set_defaults(include_knowledge_separator=True)
-        WizardDialogKnowledgeTeacher.add_cmdline_args(argparser)
-        agent = argparser.add_argument_group('GeneratorTeacher Arguments')
+    @classmethod
+    def add_cmdline_args(
+        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
+    ) -> ParlaiParser:
+        parser.set_defaults(include_knowledge_separator=True)
+        super().add_cmdline_args(parser, partial_opt=partial_opt)
+        agent = parser.add_argument_group('GeneratorTeacher Arguments')
         agent.add_argument(
             '--only-checked-knowledge',
             type='bool',
@@ -496,6 +713,7 @@ class GeneratorTeacher(WizardDialogKnowledgeTeacher):
             default='\n',
             help='delimiter for prepending gold knowledge',
         )
+        return parser
 
     def getID(self):
         return "WizTeacher"
@@ -540,6 +758,113 @@ class GeneratorTeacher(WizardDialogKnowledgeTeacher):
         return a
 
 
+class WikiPageTitleTeacher(WizardDialogKnowledgeTeacher):
+    """
+    Generates the title of Wikipedia page used as source of knowledge.
+
+    The context provided by this teacher (`text`) is the conversation history, with chosen topic removed.
+    The label is the title of the Wikipedia page of the passage that wizard selected for crafting
+    the next utterance; in other words, the source of knowledge for this utterance.
+    """
+
+    def __init__(self, opt, shared=None):
+        self.opt = copy.deepcopy(opt)
+        self.opt['label_type'] = 'response'
+        super().__init__(self.opt, shared=shared)
+        self.id = 'WikiPageTitleTeacher'
+        self._conv_history_len = self.opt['conversation_history_length']
+        if not (self._conv_history_len > 0 or self._conv_history_len == -1):
+            logging.warning(
+                f'"{self._conv_history_len}" is an invalid value for --conversation-history-length flag.'
+                ' Changing it to default of -1 (include the entire message history).'
+            )
+            self._conv_history_len = -1
+        self._skip_no_title = self.opt['skip_no_title']
+        if not shared:
+            self._preprocess_data()
+        else:
+            self.titles_data = shared['titles_data']
+
+    @classmethod
+    def add_cmdline_args(cls, parser, partial_opt=None):
+        super().add_cmdline_args(parser, partial_opt=partial_opt)
+        agent = parser.add_argument_group('Wikipedia Page Title Arguments')
+        agent.add_argument(
+            '--conversation-history-length',
+            type=int,
+            default=-1,
+            help='Number of previous utterances to keep in context, 0 (default) includes all',
+        )
+        agent.add_argument(
+            '--skip-no-title',
+            type='bool',
+            default=True,
+            help=(
+                'Whether to skip the example if no passage was selected. If `false` '
+                f'uses `{TOKEN_NOCHOSEN}` instead of title if no knowledge source was selected.'
+            ),
+        )
+        return parser
+
+    def share(self):
+        shared = super().share()
+        shared['titles_data'] = self.titles_data
+        return shared
+
+    def _generate_messages(self, hist, action):
+        include_hist = (
+            hist[-self._conv_history_len :] if self._conv_history_len > 0 else hist
+        )
+        context = '\n'.join(include_hist)
+        return Message(
+            {
+                'id': "Wikipedia Title Teacher",
+                'text': context,
+                'labels': [action["title"]],
+                'episode_done': True,
+            }
+        )
+
+    def _should_include(self, act):
+        return not (self._skip_no_title and act['labels'][0] == TOKEN_NOCHOSEN)
+
+    def _preprocess_data(self):
+        data = []
+        for episode_idx in range(super().num_episodes()):
+            dialog_history = []
+            ex_idx = 0
+            while True:
+                a = super().get(episode_idx, ex_idx)
+                text_parts = a['text'].split('\n')
+                if ex_idx == 0:
+                    # throwing away chosen_topic
+                    text_parts = text_parts[1:]
+                if text_parts:
+                    dialog_history.append(text_parts[0])
+                    title_act = self._generate_messages(dialog_history, a)
+                    if self._should_include(title_act):
+                        data.append(title_act)
+                if a['episode_done']:
+                    break
+                ex_idx += 1
+                dialog_history.append(a['labels'][0])
+
+        logging.info(
+            f'{len(data)} title generation examples generated '
+            f'from {super().num_examples()} original examples'
+        )
+        self.titles_data = data
+
+    def num_episodes(self):
+        return len(self.titles_data)
+
+    def num_examples(self):
+        return self.num_episodes()
+
+    def get(self, episode_idx, entry_idx=0):
+        return self.titles_data[episode_idx]
+
+
 ####################################################
 #                                                  #
 # Doc Reader Teachers                              #
@@ -704,10 +1029,13 @@ class DocreaderTeacher(WizardOfWikipediaTeacher):
 
         self.teacher_type = opt.get('teacher_type')
 
-    @staticmethod
-    def add_cmdline_args(argparser):
-        WizardDialogKnowledgeTeacher.add_cmdline_args(argparser)
-        argparser.add_argument(
+    @classmethod
+    def add_cmdline_args(
+        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
+    ) -> ParlaiParser:
+        super().add_cmdline_args(parser, partial_opt)
+        WizardDialogKnowledgeTeacher.add_cmdline_args(parser, partial_opt=partial_opt)
+        parser.add_argument(
             '--teacher-type',
             type=str,
             default='docs',
@@ -721,6 +1049,7 @@ class DocreaderTeacher(WizardOfWikipediaTeacher):
                 'span_teacher',
             ],
         )
+        return parser
 
     def get_min_stopwords(self, word_set):
         min_count = 1000000000000
